@@ -1,15 +1,17 @@
-#This file is part of Tryton.  The COPYRIGHT file at the top level of
-#this repository contains the full copyright notices and license terms.
+# This file is part of Tryton.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
 """
 Configuration functions for the proteus package for Tryton.
 """
-from __future__ import with_statement
-__all__ = ['set_trytond', 'set_xmlrpc', 'get_config']
-import xmlrpclib
-import threading
-from decimal import Decimal
+
 import datetime
-import time
+import os
+import threading
+import urllib.parse
+import xmlrpc.client
+from decimal import Decimal
+
+__all__ = ['set_trytond', 'set_xmlrpc', 'get_config']
 
 
 def dump_decimal(self, value, write):
@@ -19,9 +21,9 @@ def dump_decimal(self, value, write):
     self.dump_struct(value, write)
 
 
-def dump_buffer(self, value, write):
+def dump_bytes(self, value, write):
     self.write = write
-    value = xmlrpclib.Binary(value)
+    value = xmlrpc.client.Binary(value)
     value.encode(self)
     del self.write
 
@@ -44,10 +46,58 @@ def dump_time(self, value, write):
         }
     self.dump_struct(value, write)
 
-xmlrpclib.Marshaller.dispatch[Decimal] = dump_decimal
-xmlrpclib.Marshaller.dispatch[datetime.date] = dump_date
-xmlrpclib.Marshaller.dispatch[datetime.time] = dump_time
-xmlrpclib.Marshaller.dispatch[buffer] = dump_buffer
+
+def dump_timedelta(self, value, write):
+    value = {'__class__': 'timedelta',
+        'seconds': value.total_seconds(),
+        }
+    self.dump_struct(value, write)
+
+xmlrpc.client.Marshaller.dispatch[Decimal] = dump_decimal
+xmlrpc.client.Marshaller.dispatch[datetime.date] = dump_date
+xmlrpc.client.Marshaller.dispatch[datetime.time] = dump_time
+xmlrpc.client.Marshaller.dispatch[datetime.timedelta] = dump_timedelta
+if bytes != str:
+    xmlrpc.client.Marshaller.dispatch[bytes] = dump_bytes
+xmlrpc.client.Marshaller.dispatch[bytearray] = dump_bytes
+
+
+def dump_struct(self, value, write, escape=xmlrpc.client.escape):
+    converted_value = {}
+    for k, v in value.items():
+        if isinstance(k, int):
+            k = str(k)
+        elif isinstance(k, float):
+            k = repr(k)
+        converted_value[k] = v
+    return self.dump_struct(converted_value, write, escape=escape)
+
+
+xmlrpc.client.Marshaller.dispatch[dict] = dump_struct
+
+
+class XMLRPCDecoder(object):
+
+    decoders = {}
+
+    @classmethod
+    def register(cls, klass, decoder):
+        assert klass not in cls.decoders
+        cls.decoders[klass] = decoder
+
+    def __call__(self, dct):
+        if dct.get('__class__') in self.decoders:
+            return self.decoders[dct['__class__']](dct)
+        return dct
+
+XMLRPCDecoder.register('date',
+    lambda dct: datetime.date(dct['year'], dct['month'], dct['day']))
+XMLRPCDecoder.register('time',
+    lambda dct: datetime.time(dct['hour'], dct['minute'], dct['second'],
+        dct['microsecond']))
+XMLRPCDecoder.register('timedelta',
+    lambda dct: datetime.timedelta(seconds=dct['seconds']))
+XMLRPCDecoder.register('Decimal', lambda dct: Decimal(dct['decimal']))
 
 
 def end_struct(self, data):
@@ -57,18 +107,11 @@ def end_struct(self, data):
     items = self._stack[mark:]
     for i in range(0, len(items), 2):
         dct[items[i]] = items[i + 1]
-    if '__class__' in dct:
-        if dct['__class__'] == 'date':
-            dct = datetime.date(dct['year'], dct['month'], dct['day'])
-        elif dct['__class__'] == 'time':
-            dct = datetime.time(dct['hour'], dct['minute'], dct['second'],
-                dct['microsecond'])
-        elif dct['__class__'] == 'Decimal':
-            dct = Decimal(dct['decimal'])
+    dct = XMLRPCDecoder()(dct)
     self._stack[mark:] = [dct]
     self._value = 0
 
-xmlrpclib.Unmarshaller.dispatch['struct'] = end_struct
+xmlrpc.client.Unmarshaller.dispatch['struct'] = end_struct
 
 _CONFIG = threading.local()
 _CONFIG.current = None
@@ -109,6 +152,11 @@ class Config(object):
         self._context.update(kwargs)
         return ctx_manager
 
+    def reset_context(self):
+        ctx_manager = ContextManager(self)
+        self._context = {}
+        return ctx_manager
+
     def get_proxy(self, name):
         raise NotImplementedError
 
@@ -124,10 +172,11 @@ class _TrytondMethod(object):
         self._object = model
         self._config = config
 
-    def __call__(self, *args):
-        from trytond.cache import Cache
-        from trytond.transaction import Transaction
+    def __call__(self, *args, **kwargs):
         from trytond.rpc import RPC
+        from trytond.tools import is_instance_method
+        from trytond.transaction import Transaction
+        from trytond.worker import run_task
 
         if self._name in self._object.__rpc__:
             rpc = self._object.__rpc__[self._name]
@@ -138,11 +187,11 @@ class _TrytondMethod(object):
 
         with Transaction().start(self._config.database_name,
                 self._config.user, readonly=rpc.readonly) as transaction:
-            Cache.clean(self._config.database_name)
             args, kwargs, transaction.context, transaction.timestamp = \
-                rpc.convert(self._object, *args)
+                rpc.convert(self._object, *args, **kwargs)
             meth = getattr(self._object, self._name)
-            if not hasattr(meth, 'im_self') or meth.im_self:
+            if (rpc.instantiate is None
+                    or not is_instance_method(self._object, self._name)):
                 result = rpc.result(meth(*args, **kwargs))
             else:
                 assert rpc.instantiate == 0
@@ -152,9 +201,10 @@ class _TrytondMethod(object):
                 else:
                     result = [rpc.result(meth(i, *args, **kwargs))
                         for i in inst]
-            if not rpc.readonly:
-                transaction.cursor.commit()
-        Cache.resets(self._config.database_name)
+            transaction.commit()
+        while transaction.tasks:
+            task_id = transaction.tasks.pop()
+            run_task(self._config.database_name, task_id)
         return result
 
 
@@ -175,58 +225,48 @@ class TrytondProxy(object):
 class TrytondConfig(Config):
     'Configuration for trytond'
 
-    def __init__(self, database_name=None, user='admin', database_type=None,
-            language='en_US', password='', config_file=None):
+    def __init__(self, database=None, user='admin', config_file=None):
         super(TrytondConfig, self).__init__()
-        from trytond.config import CONFIG
-        CONFIG.update_etc(config_file)
-        if database_type is not None:
-            CONFIG['db_type'] = database_type
+        if not database:
+            database = os.environ.get('TRYTOND_DATABASE_URI')
+        else:
+            os.environ['TRYTOND_DATABASE_URI'] = database
+        if not config_file:
+            config_file = os.environ.get('TRYTOND_CONFIG')
+        from trytond.config import config
+        config.update_etc(config_file)
         from trytond.pool import Pool
-        from trytond import backend
-        from trytond.protocols.dispatcher import create
-        from trytond.cache import Cache
         from trytond.transaction import Transaction
-        self.database_type = CONFIG['db_type']
-        if database_name is None:
-            if self.database_type == 'sqlite':
-                database_name = ':memory:'
-            else:
-                database_name = 'test_%s' % int(time.time())
+        self.database = database
+        database_name = None
+        if database:
+            uri = urllib.parse.urlparse(database)
+            database_name = uri.path.strip('/')
+        if not database_name:
+            database_name = os.environ['DB_NAME']
         self.database_name = database_name
         self._user = user
         self.config_file = config_file
 
         Pool.start()
-
-        with Transaction().start(None, 0) as transaction:
-            cursor = transaction.cursor
-            databases = backend.get('Database').list(cursor)
-        if database_name not in databases:
-            create(database_name, CONFIG['admin_passwd'], language, password)
-
-        database_list = Pool.database_list()
         self.pool = Pool(database_name)
-        if database_name not in database_list:
-            self.pool.init()
+        self.pool.init()
 
         with Transaction().start(self.database_name, 0) as transaction:
-            Cache.clean(database_name)
             User = self.pool.get('res.user')
             transaction.context = self.context
-            self.user = User.search([
-                ('login', '=', user),
-                ], limit=1)[0].id
+            with transaction.set_context(active_test=False):
+                self.user = User.search([
+                    ('login', '=', user),
+                    ], limit=1)[0].id
             with transaction.set_user(self.user):
                 self._context = User.get_preferences(context_only=True)
-        Cache.resets(database_name)
     __init__.__doc__ = object.__init__.__doc__
 
     def __repr__(self):
-        return "proteus.config.TrytondConfig"\
-            "('%s', '%s', '%s', config_file=%s)"\
-            % (self.database_name, self._user, self.database_type,
-                self.config_file)
+        return ("proteus.config.TrytondConfig"
+            "(%s, %s, config_file=%s)"
+            % (repr(self.database), repr(self._user), repr(self.config_file)))
     __repr__.__doc__ = object.__repr__.__doc__
 
     def __eq__(self, other):
@@ -234,12 +274,12 @@ class TrytondConfig(Config):
             raise NotImplementedError
         return (self.database_name == other.database_name
             and self._user == other._user
-            and self.database_type == other.database_type
+            and self.database == other.database
             and self.config_file == other.config_file)
 
     def __hash__(self):
         return hash((self.database_name, self._user,
-            self.database_type, self.config_file))
+            self.database, self.config_file))
 
     def get_proxy(self, name, type='model'):
         'Return Proxy class'
@@ -254,11 +294,10 @@ class TrytondConfig(Config):
         return methods
 
 
-def set_trytond(database_name=None, user='admin', database_type=None,
-        language='en_US', password='', config_file=None):
+def set_trytond(database=None, user='admin',
+        config_file=None):
     'Set trytond package as backend'
-    _CONFIG.current = TrytondConfig(database_name, user, database_type,
-            language=language, password=password, config_file=config_file)
+    _CONFIG.current = TrytondConfig(database, user, config_file=config_file)
     return _CONFIG.current
 
 
@@ -279,17 +318,18 @@ class XmlrpcProxy(object):
 class XmlrpcConfig(Config):
     'Configuration for XML-RPC'
 
-    def __init__(self, url):
+    def __init__(self, url, **kwargs):
         super(XmlrpcConfig, self).__init__()
         self.url = url
-        self.server = xmlrpclib.ServerProxy(url, allow_none=1, use_datetime=1)
+        self.server = xmlrpc.client.ServerProxy(
+            url, allow_none=1, use_datetime=1, **kwargs)
         # TODO add user
         self.user = None
         self._context = self.server.model.res.user.get_preferences(True, {})
     __init__.__doc__ = object.__init__.__doc__
 
     def __repr__(self):
-        return "proteus.config.XmlrpcConfig('%s')" % self.url
+        return "proteus.config.XmlrpcConfig(%s)" % repr(self.url)
     __repr__.__doc__ = object.__repr__.__doc__
 
     def __eq__(self, other):
@@ -313,9 +353,12 @@ class XmlrpcConfig(Config):
                 and '.' not in x[len(object_) + 1:]]
 
 
-def set_xmlrpc(url):
-    'Set XML-RPC as backend'
-    _CONFIG.current = XmlrpcConfig(url)
+def set_xmlrpc(url, **kwargs):
+    '''
+    Set XML-RPC as backend.
+    It pass the keyword arguments received to xmlrpclib.ServerProxy()
+    '''
+    _CONFIG.current = XmlrpcConfig(url, **kwargs)
     return _CONFIG.current
 
 
